/* File:		maths.cpp */
/* Description:	This file module contains maths support functions */

#include <math.h>
#include "types.h"
#include "maths.h"

/* Function:    sind */
/* Description: This function returns the sine of an angle specified in degrees. */

float sind(float angle)
{
  return sin(angle/RADS_TO_DEGS);
}

/* Function:    cosd */
/* Description: This function returns the cosine of an angle specified in degrees. */

float cosd(float angle)
{
  return cos(angle/RADS_TO_DEGS);
}

/* Function:    acosd */
/* Description: This function returns an angle in degrees calculated from a */
/*				constant equal to the cosine of the angle. */

float acosd(float c)
{
  return (acos(c) * RADS_TO_DEGS);
}

/* Function:    world_vector_add */
/* Description: This function returns the result of a vector addition using */
/*              the two supplied vectors. */

world_vector world_vector_add(world_vector a, world_vector b)
{
  world_vector c;     /* temp store of result */

  c.x = a.x + b.x;
  c.y = a.y + b.y;
  c.z = a.z + b.z;

  return c;
}

/* Function:    world_vector_subtract */
/* Description: This function returns the result of a vector subtraction using */
/*              the two supplied vectors. */

world_vector world_vector_subtract(world_vector a, world_vector b)
{
  world_vector c;     /* temp store of result */

  c.x = a.x - b.x;
  c.y = a.y - b.y;
  c.z = a.z - b.z;

  return c;
}

/* Function:    world_vector_scale */
/* Description: This function returns the supplied vector multiplied */
/*              by the supplied scaler value. */

world_vector world_vector_scale(world_vector a, float scale)
{
  world_vector b;     /* temp store of result */

  b.x = a.x * scale;
  b.y = a.y * scale;
  b.z = a.z * scale;

  return b;
}

/* Function:    world_vector_magnitude */
/* Description: This function returns the magnitude of the supplied vector. */

float world_vector_magnitude(world_vector a)
{
  return sqrt(pow(a.x, 2) + pow(a.y, 2) + pow(a.z, 2));
}

/* Function:    calculate_unit_vector */
/* Description: This function calculates a unit vector using */
/*              the two supplied vector & its magnitude */

world_vector calculate_unit_vector(world_vector v1, float mag)
{
  world_vector v;     /* temp store of result */

  if (mag != 0)
  {
	v.x = v1.x / mag;
	v.y = v1.y / mag;
	v.z = v1.z / mag;
  }
  else
  {
	v.x = 0.0;
	v.y = 0.0;
	v.z = 0.0;
  }

  return v;
}

/* Function:    world_vector_cross */
/* Description: This function returns the result of a vector cross product using */
/*              the two supplied vectors. */

world_vector world_vector_cross(world_vector a, world_vector b)
{
  world_vector c;     /* temp store of result */

  c.x = a.y * b.z - a.z * b.y;
  c.y = a.z * b.x - a.x * b.z;
  c.z = a.x * b.y - a.y * b.x;

  return c;
}

/* Function:    world_vector_dot */
/* Description: This function returns the result of a vector dot product using the */
/*              two supplied vectors. */

float world_vector_dot(world_vector a, world_vector b)
{
  return (a.z * b.z) + (a.x * b.x) + (a.y * b.y);
}

/* Function:    calculate_panel_normal */
/* Description: This function returns the panel normal calculated */
/*              from the three supplied vectices (the vertices MUST */
/*              be in clockwise order). */

world_vector calculate_panel_normal(world_vector v1,
									world_vector v2,
									world_vector v3)
{
	world_vector n;
	world_vector a, b;
	float theta;

	a = world_vector_subtract(v2, v1);
	b = world_vector_subtract(v3, v2);
	a = calculate_unit_vector(a, world_vector_magnitude(a));
	b = calculate_unit_vector(b, world_vector_magnitude(b));
	theta = acosd(world_vector_dot(a, b));
	n = world_vector_cross(a, b);
	n.x = n.x / sind(theta);
	n.y = n.y / sind(theta);
	n.z = n.z / sind(theta);

	return n;
}

/* Function:    calculate_dir_cosines */
/* Description: This function returns the direction cosines calculated from the */
/*              supplied angles. */

dir_cosine calculate_dir_cosines(orientation angle)
{
  float sin_pitch, cos_pitch;
  float sin_yaw, cos_yaw;
  float sin_roll, cos_roll;
  dir_cosine dc;

  sin_pitch = sind(-angle.pitch);
  cos_pitch = cosd(-angle.pitch);
  sin_yaw = sind(angle.yaw);
  cos_yaw = cosd(angle.yaw);
  sin_roll = sind(angle.roll);
  cos_roll = cosd(angle.roll);

  dc.l1 = cos_pitch * cos_yaw;
  dc.l2 = cos_pitch * sin_yaw;
  dc.l3 = -sin_pitch;

  dc.m1 = (sin_pitch * sin_roll * cos_yaw) - (cos_roll * sin_yaw);
  dc.m2 = (sin_pitch * sin_roll * sin_yaw) + (cos_roll * cos_yaw);
  dc.m3 = cos_pitch * sin_roll;

  dc.n1 = (sin_pitch * cos_roll * cos_yaw) + (sin_roll * sin_yaw);
  dc.n2 = (sin_pitch * cos_roll * sin_yaw) - (sin_roll * cos_yaw);
  dc.n3 = cos_pitch * cos_roll;

  return dc;
}

/* Function:    transform_object_to_world */
/* Description: This function applies the supplied direction cosines to transform */
/*              the object co-ordinates to world co-ordinates. */

world_vector transform_object_to_world(world_vector object, dir_cosine dc)
{
  world_vector world;      /* temp store of result */

  world.z = (dc.l1 * object.z) + (dc.m1 * object.x) + (dc.n1 * object.y);
  world.x = (dc.l2 * object.z) + (dc.m2 * object.x) + (dc.n2 * object.y);
  world.y = (dc.l3 * object.z) + (dc.m3 * object.x) + (dc.n3 * object.y);

  return world;
}

/* Function:    orientation_to_unit_vector */
/* Description: This function converts the supplied orientation to its */
/*				equivalent unit direction vector */

world_vector orientation_to_unit_vector(orientation o)
{
	world_vector v;
	dir_cosine dc;

	dc = calculate_dir_cosines(o);
	v.x = 0.0; v.y = 0.0; v.z = 1.0;
	v = transform_object_to_world(v, dc);

	return v;
}

