/* File:		terrain.c */
/* Description:	This file module contains the terrain loading and rendering functions */

#include <stdio.h>
#include <GL/glut.h>
#include "types.h"
#include "maths.h"
#include "terrain.h"

/* variables */

static unsigned char heightmap[257][257];
static colour_rgb colourmap[257][257];

/* Function to calculate a colour value for a vertex */

colour_rgb calculate_vertex_colour(unsigned char height)
{
	colour_rgb c;
	float alpha;

	if (height <= WATER_HEIGHT)
	{
		c.r = WATER_COLOUR_R;
		c.g = WATER_COLOUR_G;
		c.b = WATER_COLOUR_B;
	} else if (height < SHORE_HEIGHT)
	{
		alpha = (float)(height - WATER_HEIGHT) / (float)(SHORE_HEIGHT - WATER_HEIGHT);
		c.r = ((1.0 - alpha) * WATER_COLOUR_R) + (alpha * GRASS_COLOUR_R);
		c.g = ((1.0 - alpha) * WATER_COLOUR_G) + (alpha * GRASS_COLOUR_G);
		c.b = ((1.0 - alpha) * WATER_COLOUR_B) + (alpha * GRASS_COLOUR_B);
	} else if (height < GRASS_HEIGHT)
	{
		c.r = GRASS_COLOUR_R;
		c.g = GRASS_COLOUR_G;
		c.b = GRASS_COLOUR_B;
	} else if (height < ROCK_HEIGHT)
	{
		alpha = (float)(height - GRASS_HEIGHT) / (float)(ROCK_HEIGHT - GRASS_HEIGHT);
		c.r = ((1.0 - alpha) * GRASS_COLOUR_R) + (alpha * SNOW_COLOUR_R);
		c.g = ((1.0 - alpha) * GRASS_COLOUR_G) + (alpha * SNOW_COLOUR_G);
		c.b = ((1.0 - alpha) * GRASS_COLOUR_B) + (alpha * SNOW_COLOUR_B);
	} else
	{
		c.r = SNOW_COLOUR_R;
		c.g = SNOW_COLOUR_G;
		c.b = SNOW_COLOUR_B;
	}

	return c;
}

/* Function to calculate the lighting for the vertices */

void calculate_vertex_lighting()
{
	world_vector panel_normals[256][256];	/* array of normals for each quad in scene */
	world_vector v1, v2, v3;			/* temporary vertex stores */
	world_vector n;						/* current vertex normal */
	int num_normals;					/* number of quad normals that apply to a vertex normal */
	world_vector lv;					/* direction of light source */
	float intensity;					/* calculated light intensity for current vertex */
	int x, z;							/* counters */

	/* calculate the normal for each quad panel in the scene */

	for (x=0; x<256; x++)
	{
		for (z=0; z<256; z++)
		{
			v1.x = x*GRID_FACTOR; v1.y = heightmap[x][z]*HEIGHT_FACTOR; v1.z = z*GRID_FACTOR;
			v2.x = x*GRID_FACTOR; v2.y = heightmap[x][z+1]*HEIGHT_FACTOR; v2.z = (z+1)*GRID_FACTOR;
			v3.x = (x+1)*GRID_FACTOR; v3.y = heightmap[x+1][z+1]*HEIGHT_FACTOR; v3.z = (z+1)*GRID_FACTOR;
			panel_normals[x][z] = calculate_panel_normal(v1, v2, v3);
		}
	}

	/* average the panel normals at each vertex and calculate the lighting */

	for (x=0; x<257; x++)
	{
		for (z=0; z<257; z++)
		{
			num_normals = 0;
			n.x = 0.0; n.y = 0.0; n.z = 0.0;
			if (x>0 && z<256)
			{
				n.x = n.x + panel_normals[x-1][z].x;
				n.y = n.y + panel_normals[x-1][z].y;
				n.z = n.z + panel_normals[x-1][z].z;
				num_normals++;
			}
			if (x<256 && z<256)
			{
				n.x = n.x + panel_normals[x][z].x;
				n.y = n.y + panel_normals[x][z].y;
				n.z = n.z + panel_normals[x][z].z;
				num_normals++;
			}
			if (x<256 && z>0)
			{
				n.x = n.x + panel_normals[x][z-1].x;
				n.y = n.y + panel_normals[x][z-1].y;
				n.z = n.z + panel_normals[x][z-1].z;
				num_normals++;
			}
			if (x>0 && z>0)
			{
				n.x = n.x + panel_normals[x-1][z-1].x;
				n.y = n.y + panel_normals[x-1][z-1].y;
				n.z = n.z + panel_normals[x-1][z-1].z;
				num_normals++;
			}

			/* calculate vertex normal */

			n = world_vector_scale(n, (1.0/(float)num_normals));

			/* apply lighting */

			lv.x = LIGHT_ANGLE_X; lv.y = LIGHT_ANGLE_Y; lv.z = LIGHT_ANGLE_Z;
			intensity = -world_vector_dot(n, lv);
			if (intensity < 0.0)
			{
				intensity = 0.0;
			}
			colourmap[x][z].r = colourmap[x][z].r * (AMBIENT_LIGHT_R + (intensity * LIGHT_COLOUR_R));
			colourmap[x][z].g = colourmap[x][z].g * (AMBIENT_LIGHT_G + (intensity * LIGHT_COLOUR_G));
			colourmap[x][z].b = colourmap[x][z].b * (AMBIENT_LIGHT_B + (intensity * LIGHT_COLOUR_B));
		}
	}

}

/* Function to load and create a terrain database from a bmp file */

void make_terrain(char filename[])
{
	long data_start = 54;		/* start of data section in terrain .bmp files */
								/* NB header size of bmp files is not fixed, refer to M$ documentation */
	unsigned char ch, cho[3];	/* temp file input buffers */
	FILE *fileptr, *fileptr2;				/* pointer to bmp file */
	int n, m, i, j;				/* counters */

	/* open file */

	if ((fileptr = fopen(filename, "rb")) == NULL)
		return;

	/* step over the header */

	n = 0;
	while (!feof(fileptr) && n < data_start)
	{
		ch = fgetc(fileptr);
		n++;
	}

	/* load the height map data */

	for (i=0; i<257; i++)
	{
		for (j=0; j<257; j++)
		{
			cho[0] = fgetc(fileptr);
			cho[1] = fgetc(fileptr);
			cho[2] = fgetc(fileptr);
			heightmap[j][i] = cho[0];
		}
		cho[0] = fgetc(fileptr);	/* remove padding byte */
	}

	/* close the file */

	fclose(fileptr);

	/* generate the colour map data */

	for (i=0; i<257; i++)
	{
		for (j=0; j<257; j++)
		{
			colourmap[i][j] = calculate_vertex_colour(heightmap[i][j]);
		}
	}

	/* light the scene */

	calculate_vertex_lighting();
}

/* Function to render terrain */

void render_terrain()
{
	int s, t;

	/* render terrain */

	for (t=0; t<256; t++)
	{
		glBegin(GL_TRIANGLE_STRIP);
			glColor3f(colourmap[0][t+1].r, colourmap[0][t+1].g, colourmap[0][t+1].b);
			glTexCoord2f(0.0, (float)(t+1)*TEX_SCALE);
			glVertex3f(0.0, (heightmap[0][t+1]*HEIGHT_FACTOR), -((t+1)*GRID_FACTOR));
			glColor3f(colourmap[0][t].r, colourmap[0][t].g, colourmap[0][t].b);
			glTexCoord2f(0.0, (float)t*TEX_SCALE);
			glVertex3f(0.0, (heightmap[0][t]*HEIGHT_FACTOR), -(t*GRID_FACTOR));
			for (s=1; s<257; s++)
			{
				glColor3f(colourmap[s][t+1].r, colourmap[s][t+1].g, colourmap[s][t+1].b);
				glTexCoord2f((float)s*TEX_SCALE, (float)(t+1)*TEX_SCALE);
				glVertex3f((s*GRID_FACTOR), (heightmap[s][t+1]*HEIGHT_FACTOR), -((t+1)*GRID_FACTOR));
				glColor3f(colourmap[s][t].r, colourmap[s][t].g, colourmap[s][t].b);
				glTexCoord2f((float)s*TEX_SCALE, (float)t*TEX_SCALE);
				glVertex3f((s*GRID_FACTOR), (heightmap[s][t]*HEIGHT_FACTOR), -(t*GRID_FACTOR));
			}
		glEnd();
	}
}
